# Default Python Libraries
import asyncio
from asyncio.base_events import BaseEventLoop
from multiprocessing.pool import ThreadPool

from .client_service import AristaFlowClientService
from .configuration import Configuration
from .rest_helper import RestPackageRegistry
from .service_provider import ServiceProvider


class AristaFlowClientPlatform(object):
    """Entry point to the AristaFlow Python Client framework."""

    # thread pool for async requests
    __async_thread_pool: ThreadPool = None
    __push_event_loop: BaseEventLoop = None

    def __init__(self, configuration: Configuration):
        self.configuration = configuration
        self.__client_services: [str, AristaFlowClientService] = {}
        self.__rest_package_registry = RestPackageRegistry(configuration)
        self.__async_thread_pool = ThreadPool(configuration.async_thread_pool_size)
        self.__push_event_loop = asyncio.new_event_loop()
        self.__async_thread_pool.apply_async(self._start_event_loop)

    def get_client_service(self, user_session: str = "python_default_session"):
        """
        :return: AristaFlowClientService The client service for the given user session
        """
        if user_session in self.__client_services:
            return self.__client_services[user_session]
        cs = AristaFlowClientService(
            self.configuration,
            user_session,
            ServiceProvider(
                self.__rest_package_registry, self.__async_thread_pool, self.__push_event_loop
            ),
        )
        self.__client_services[user_session] = cs
        return cs

    def _start_event_loop(self):
        """
        Starts the asyncio event loop for handling push notifications
        """
        try:
            asyncio.set_event_loop(self.__push_event_loop)
            self.__push_event_loop.run_forever()
        finally:
            self.__push_event_loop.close()



# -*- coding: utf-8 -*-
# Default Python Libraries
import asyncio
import json
import traceback
from asyncio import sleep
from typing import List, Optional, Set, Union

# AristaFlow REST Libraries
from af_execution_manager import InstanceCreationSseData, InstanceStateData
from af_execution_manager.api.instance_control_api import InstanceControlApi
from af_execution_manager.api.templ_ref_remote_iterator_rest_api import (
    TemplRefRemoteIteratorRestApi,
)
from af_execution_manager.models.data_container import DataContainer
from af_execution_manager.models.instance_creation_data import InstanceCreationData
from af_execution_manager.models.instance_creation_rest_data import InstanceCreationRestData
from af_execution_manager.models.parameter_value import ParameterValue
from af_execution_manager.models.templ_ref_initial_remote_iterator_data import (
    TemplRefInitialRemoteIteratorData,
)
from af_execution_manager.models.templ_ref_remote_iterator_data import TemplRefRemoteIteratorData
from af_execution_manager.models.template_reference import TemplateReference
from af_process_manager.api.instance_manager_api import InstanceManagerApi
from af_process_manager.models.instance_reference import InstanceReference
from aristaflow.abstract_service import AbstractService
from aristaflow.configuration import Configuration
from aristaflow.service_provider import ServiceProvider
from aristaflow.utils import VERSION


def __instance_state_listener(instance_state: InstanceStateData):
    pass


InstanceStateListener = type(__instance_state_listener)
""" Type for callback listeners"""


class ProcessService(AbstractService):
    """Process related methods"""

    __instance_state_listeners: Set[InstanceStateListener] = None
    __push_sse_client = None
    _sse_id: str = None
    __af_conf: Configuration = None

    def __init__(self, service_provider: ServiceProvider, configuration: Configuration):
        self.__instance_state_listeners = set()
        self.__af_conf = configuration
        super().__init__(service_provider=service_provider)

    def __tpl_version_key(self, tpl: TemplateReference) -> int:
        return VERSION.key(tpl.version)

    def get_instantiable_template_by_type(self, process_type: str) -> Optional[TemplateReference]:
        """Finds the first instantiable template of the given process Type"""
        tpls = self.get_instantiable_templates()
        tpls_of_type = []
        for tpl in tpls:
            if tpl.process_type == process_type:
                tpls_of_type += [tpl]
        # empty result

        if len(tpls_of_type) == 0:
            return None
        # reverse sort by version
        tpls_of_type = sorted(tpls_of_type, key=self.__tpl_version_key, reverse=True)
        # return first entry
        return tpls_of_type[0]

    def get_instantiable_templates(self) -> List[TemplateReference]:
        """Retrieves the instantiable tempaltes from the server
        :return: List[TemplateReference] The instantiable templates for the current user
        """
        ic: InstanceControlApi = self._service_provider.get_service(InstanceControlApi)
        initial: TemplRefInitialRemoteIteratorData = ic.get_instantiable_templ_refs()
        tpls: List[TemplateReference] = []
        self.__iterate(tpls, initial)
        return tpls

    def __iterate(
        self,
        tpls: List[TemplateReference],
        inc: Union[TemplRefInitialRemoteIteratorData, TemplRefRemoteIteratorData],
    ):
        """Consumes an template reference remote iterator until it is used up
        @param tpls The tpls list to fill with the template references
        @param inc The first or next iteration to consume and append to tpls.
        """
        if inc is None:
            return
        # append the tpls
        if inc.templ_refs:
            tpls += inc.templ_refs
        else:
            return
        # iterator is used up
        if inc.dropped:
            return

        # fetch next
        tref_rest: TemplRefRemoteIteratorRestApi = self._service_provider.get_service(
            TemplRefRemoteIteratorRestApi
        )
        next_it: TemplRefRemoteIteratorData = tref_rest.templ_ref_get_next(inc.iterator_id)
        self.__iterate(tpls, next_it)

    def start_by_type(
        self, process_type: str, callback_uri: str = None, input_data: dict = None
    ) -> str:
        """Starts the newest version of the given process type, returns the logical ID of the started instance."""
        tpl = self.get_instantiable_template_by_type(process_type)
        if tpl is None:
            raise Exception("Unknown process type: " + process_type)
        return self.start_by_id(tpl.id, callback_uri, input_data)

    def start_by_id(
        self, template_id: str, callback_uri: str = None, input_data: dict = None
    ) -> str:
        """Starts a process given by the template id. Returns the logical ID of the started instance.
        If an InstanceStateListener is registered, instance state changes (and autostart pending events) are
        listened to using SSE. In that case, no callback_uri can be used.
        """
        ic: InstanceControlApi = self._service_provider.get_service(InstanceControlApi)
        if callback_uri is not None and self._sse_id is not None:
            raise Exception("Can not use Callbacks and SSE listener at the same time")
        elif self._sse_id is not None:
            inst_creation_data = InstanceCreationSseData(
                sub_class="InstanceCreationSseData", sse_conn=self._sse_id
            )
            inst_creation_data.dc = self.__create_instance_container(ic, template_id, input_data)
            return ic.create_and_start_instance_sse(templ_id=template_id, body=inst_creation_data)
        elif callback_uri is not None:
            inst_creation_data = InstanceCreationRestData(
                sub_class="InstanceCreationRestData", notification_callback=callback_uri
            )
            inst_creation_data.dc = self.__create_instance_container(ic, template_id, input_data)
            return ic.create_and_start_instance_callback(
                body=inst_creation_data, templ_id=template_id
            )
        else:
            # no callback url, no SSE
            inst_creation_data = InstanceCreationData(sub_class="InstanceCreationData")
            inst_creation_data.dc = self.__create_instance_container(ic, template_id, input_data)
            return ic.create_and_start_instance(template_id, body=inst_creation_data)

    def __create_instance_container(
        self, ic: InstanceControlApi, template_id: str, input_data: dict
    ) -> DataContainer:
        """Creates an instance data container for the given template, if required"""
        idc = None
        if input_data is not None and len(input_data) != 0:
            idc = ic.create_instance_data_container(template_id)
            pv: ParameterValue
            for pv in idc.values:
                if pv.name in input_data:
                    value = input_data[pv.name]
                    pv.value = value
        return idc

    def _enable_sse_instance_state_notification(self):
        """
        Enable propagation of instance state changes using SSE push notifications.
        """
        if self.__push_sse_client is not None:
            return
        # print("ProcessService: Enabling SSE")
        asyncio.run_coroutine_threadsafe(
            self._process_push_instance_state_changes(), self._service_provider.push_event_loop
        )

        # Default Python Libraries
        import time

        # wait and check for sse_id to be set
        i = 0
        while i < 20 and self._sse_id is None:
            time.sleep(0.1)
            i = i + 1
        # print(f"SSE available after {i/10} seconds  ID is {self._sse_id}")

    async def _process_push_instance_state_changes(self):
        """
        Coroutine retrieving SSE push notifications instance state changes
        """
        while True:
            # print("Establishing SSE connection...")
            try:
                sse_connection_id, sse_client = self._service_provider.connect_sse(
                    InstanceControlApi
                )
                self._sse_id = sse_connection_id
                # print(f'ProcessService: SSE connection establised, id {self._sse_id}')
                while True:
                    self.__push_sse_client = sse_client
                    for event in sse_client:
                        if event.event == "SseConnectionEstablished":
                            print("ProcessService SSE session was re-established.")
                            self._sse_id = event.data
                            # callback_data.sse_conn = event.data
                        else:
                            # print("Instance state data received")
                            try:
                                instance_state_dict = json.loads(event.data)
                                instance_state: InstanceStateData = (
                                    self._service_provider.deserialize(
                                        instance_state_dict, InstanceStateData
                                    )
                                )
                                # call the listeners
                                self._notify_instance_state_listeners(instance_state)
                            except Exception as e:
                                print("Couldn't deserialize and apply update: ", event, e)
            #                        else:
            #                            print(f"Unknown worklist SSE push event {event.event} received")
            except ConnectionError:
                # re-establish connection after some wait time
                # print("ProcessService SSE disconnected...")
                await sleep(self.__af_conf.sse_connect_retry_wait)
            except Exception as e:
                print("Unknown exception caught during SSE handling", e.__class__)
                raise
            finally:
                print("ProcessService: SSE disconnected")
                self.__push_sse_client = None
                self._sse_id = None

    def add_instance_state_listener(self, listener: InstanceStateListener):
        """
        Adds a listener which is called whenever an instance state change event occurs. This only applies for instances
        started by this service and enables SSE.
        """
        self._enable_sse_instance_state_notification()
        self.__instance_state_listeners.add(listener)

    def remove_instance_state_listener(self, listener: InstanceStateListener):
        """
        Removes the given instance state listener
        """
        self.__instance_state_listeners.remove(listener)

    def _notify_instance_state_listeners(self, instance_state: InstanceStateData):
        """
        Notifies all registered instance state listeners
        """
        for listener in self.__instance_state_listeners:
            try:
                listener(instance_state)
            except Exception as e:
                print("Caught exception while notifying listener:", e)
                traceback.print_exc()

    def get_instance_ref(self, inst_id: str) -> InstanceReference:
        """
        Finds the instance reference for the given instance ID (logical or log ID)
        """
        im: InstanceManagerApi = self._service_provider.get_service(InstanceManagerApi)
        try:
            return im.get_instance_refs(body=[inst_id]).inst_refs[0]
        except Exception:
            logicl_id = im.get_logical_instance_ids(body=[inst_id]).inst_ids[0]
            return im.get_instance_refs(body=[logicl_id]).inst_refs[0]



# -*- coding: utf-8 -*-
# AristaFlow REST Libraries
from aristaflow.remote_iterator_handler import RemoteIteratorHandler
from aristaflow.service_provider import ServiceProvider


class AbstractService(object):
    """Abstract base class for service helpers"""

    _service_provider: ServiceProvider = None
    _rem_iter_handler: RemoteIteratorHandler = None

    def __init__(self, service_provider: ServiceProvider):
        self._service_provider = service_provider
        self._rem_iter_handler = RemoteIteratorHandler(service_provider)



# -*- coding: utf-8 -*-
# Default Python Libraries
import asyncio
import json
import traceback
import warnings
from asyncio import sleep
from typing import Generator, List, Set, Union

# Third Party Libraries
from requests import ConnectionError

# AristaFlow REST Libraries
from af_worklist_manager.api.inc_client_worklists_api import IncClientWorklistsApi
from af_worklist_manager.api.inc_worklist_update_api import IncWorklistUpdateApi
from af_worklist_manager.api.worklist_update_manager_api import WorklistUpdateManagerApi
from af_worklist_manager.models.client_worklist_item import ClientWorklistItem
from af_worklist_manager.models.client_worklist_item_update import ClientWorklistItemUpdate
from af_worklist_manager.models.client_worklist_rest_callback_data import (
    ClientWorklistRestCallbackData,
)
from af_worklist_manager.models.client_worklist_sse_callback_data import (
    ClientWorklistSseCallbackData,
)
from af_worklist_manager.models.inc_client_worklist_data import IncClientWorklistData
from af_worklist_manager.models.inc_worklist_update_data import IncWorklistUpdateData
from af_worklist_manager.models.initial_inc_client_worklist_data import InitialIncClientWorklistData
from af_worklist_manager.models.initial_inc_worklist_update_data import InitialIncWorklistUpdateData
from af_worklist_manager.models.update_interval import UpdateInterval
from af_worklist_manager.models.worklist_revision import WorklistRevision
from af_worklist_manager.models.worklist_update import WorklistUpdate
from af_worklist_manager.models.worklist_update_configuration import WorklistUpdateConfiguration

from .configuration import Configuration
from .service_provider import ServiceProvider
from .worklist_model import Worklist


# signature for worklist updates
def __update_listener(updates: List[ClientWorklistItemUpdate]):
    pass


WorklistUpdateListener = type(__update_listener)
"""
    Type for callback function for retrieving worklist updates
"""


class WorklistService(object):
    # The fetch size for incremental worklists / updates. If None, the
    # server-side default will be used
    fetch_count: int = None
    __worklist: Worklist = None
    __items: List[ClientWorklistItem] = None
    __service_provider: ServiceProvider = None
    __worklist_callback: str = None
    __worklist_update_listeners: Set[WorklistUpdateListener] = None
    __af_conf: Configuration = None
    __push_sse_client = None

    def __init__(self, service_provider: ServiceProvider, configuration: Configuration):
        self.__items = []
        self.__service_provider = service_provider
        self.__worklist_update_listeners = set()
        self.__af_conf = configuration

    def create_worklist_update_configuration(self, push: bool) -> WorklistUpdateConfiguration:
        """Creates a default worklist update configuration
        :param bool push: If set to true, create a worklist update configuration for push notifications.
        """
        update_intervals: list[UpdateInterval] = []
        if push:
            update_intervals.append(UpdateInterval(0, 200))

        # worklistFilter: NO_TL or TL_ONLY
        # notify_only: if set to true, SSE push will send notifications instead of the updates themselves
        wuc = WorklistUpdateConfiguration(
            update_mode_threshold=0,
            update_intervals=update_intervals,
            worklist_filter="NO_TL",
            notify_only=False,
        )
        return wuc

    def get_worklist(self, worklist_callback: str = None) -> List[ClientWorklistItem]:
        """
        Updates and returns the worklist of the current user
        :param str worklist_callback: Optionally an URL implementing the callback endpoint for push updates
        """
        if self.__worklist is not None:
            # simply return the items if push notifications are enabled
            if self.__worklist_callback is not None:
                return self.__items
            # perform update
            return self.update_worklist()

        wum: WorklistUpdateManagerApi = self.__service_provider.get_service(
            WorklistUpdateManagerApi
        )
        update_conf: WorklistUpdateConfiguration = self.create_worklist_update_configuration(True)
        wlit: InitialIncClientWorklistData = None
        if self.fetch_count is not None:
            wlit = wum.logon_and_create_client_worklist(body=update_conf, count=self.fetch_count)
        else:
            wlit = wum.logon_and_create_client_worklist(body=update_conf)

        # currently no items in the worklist
        if wlit is None:
            return self.__items

        self.__iterate(self.__items, wlit)

        # remember the current worklist meta data
        self.__worklist = Worklist(
            wlit.worklist_id, wlit.revision, wlit.client_worklist_id, update_conf, wlit.agent
        )

        if worklist_callback is not None:
            callbackData = ClientWorklistRestCallbackData(
                worklist_callback=worklist_callback,
                sub_class="ClientWorklistRestCallbackData",
                id=self.__worklist.worklist_id,
                client_worklist_id=self.__worklist.client_worklist_id,
                agent=self.__worklist.agent,
                revision=self.__worklist.revision,
                wu_conf=self.__worklist.wu_conf,
            )
            wum.register_client_worklist_callback(callbackData)
            self.__worklist_callback = worklist_callback

        return self.__items

    def enable_push_updates(self):
        """
        Enable automatic worklist updates using SSE push notifications.
        """
        if self.__push_sse_client is not None:
            return
        # ensure the worklist has been fetched once
        # NOTE: this does not help, if the worklist is empty (BPM-3581)
        if self.__worklist is None:
            self.get_worklist()
        if self.__worklist is None:
            warnings.warn(
                "The worklist could not be initialized, probably due to BPM-3581. SSE push will not work."
            )
        asyncio.run_coroutine_threadsafe(
            self._process_push_updates(), self.__service_provider.push_event_loop
        )

    async def _process_push_updates(self):
        """
        Coroutine retrieving SSE push notifications for the worklist, handling registration and reconnects
        """
        while True:
            # print("Establishing SSE connection...")
            try:
                sse_connection_id, sse_client = self.__service_provider.connect_sse(
                    WorklistUpdateManagerApi
                )
                while True:
                    # print("SSE connection established, registering for worklist push...")
                    callback_data = ClientWorklistSseCallbackData(
                        sse_conn=sse_connection_id,
                        sub_class="ClientWorklistSseCallbackData",
                        id=self.__worklist.worklist_id,
                        client_worklist_id=self.__worklist.client_worklist_id,
                        agent=self.__worklist.agent,
                        revision=self.__worklist.revision,
                        wu_conf=self.__worklist.wu_conf,
                    )
                    wum: WorklistUpdateManagerApi = self.__service_provider.get_service(
                        WorklistUpdateManagerApi
                    )
                    wum.register_client_worklist_sse(callback_data)
                    # print("Worklist registered for SSE push")
                    self.__push_sse_client = sse_client
                    for event in sse_client:
                        if event.event == "SseConnectionEstablished":
                            # print('SSE session was re-established, re-registering..')
                            callback_data.sse_conn = event.data
                            callback_data.revision = (self.__worklist.revision,)
                            wum.register_client_worklist_sse(callback_data)
                            # print("Worklist registered again for SSE push")
                        elif event.event == "client-worklist-update":
                            # print("Worklist update received")
                            try:
                                update_dict = json.loads(event.data)
                                update: WorklistUpdate = self.__service_provider.deserialize(
                                    update_dict, WorklistUpdate
                                )
                                self.__apply_worklist_updates(
                                    update.source_revision,
                                    update.target_revision,
                                    update.item_updates,
                                )
                                # call the listeners
                                self._notify_worklist_update_listeners(update.item_updates)
                            except Exception as e:
                                print("Couldn't deserialize and apply update: ", event, e)
                        else:
                            print(f"Unknown worklist SSE push event {event.event} received")
            except ConnectionError:
                # re-establish connection after some wait time
                # print("SSE disconnected...")
                await sleep(self.__af_conf.sse_connect_retry_wait)
            except Exception as e:
                print("Unknown exception caught during SSE handling", e.__class__)
                traceback.print_exc()
                raise
            finally:
                self.__push_sse_client = None

    def add_update_listener(self, listener: WorklistUpdateListener):
        """
        Adds a listener which is called after a worklist update was received and applied.
        """
        self.__worklist_update_listeners.add(listener)

    def remove_update_listener(self, listener: WorklistUpdateListener):
        """
        Removes the given worklist update listener
        """
        self.__worklist_update_listeners.remove(listener)

    def _notify_worklist_update_listeners(self, updates: List[ClientWorklistItemUpdate]):
        """
        Notifies all registered worklist update listeners
        """
        for listener in self.__worklist_update_listeners:
            try:
                listener(updates)
            except Exception as e:
                print("Caught exception while notifying listener:", e)
                traceback.print_exc()

    def worklist_meta_data(self) -> Worklist:
        """
        Returns the worklist meta data, like ID, current revision etc., don't modify.
        """
        if self.__worklist is None:
            self.get_worklist()
        return self.__worklist

    def __iterate(
        self,
        items: List[ClientWorklistItem],
        inc: Union[InitialIncClientWorklistData, IncClientWorklistData],
    ):
        """Consumes an incremental client worklist until its iterator is used up
        @param items The items list to fill with the update(s)
        @param inc The first or next iteration to consume and append to items.
        """
        if inc is None:
            return
        # append the items
        if inc.items_flat:
            items += inc.items_flat
        else:
            return
        # iterator is used up
        if inc.dropped:
            return

        # fetch next
        inc_cl: IncClientWorklistsApi = self.__service_provider.get_service(IncClientWorklistsApi)
        next_it: IncClientWorklistData = inc_cl.inc_client_wl_get_next(inc.inc_wl_id)
        self.__iterate(items, next_it)

    def update_worklist(self) -> List[ClientWorklistItem]:
        """Updates the user's worklist and returns the items"""
        if self.__worklist is None:
            return self.get_worklist()

        if self.__push_sse_client is not None:
            return self.__items

        wu: WorklistUpdateManagerApi = self.__service_provider.get_service(WorklistUpdateManagerApi)
        inc_updts: InitialIncWorklistUpdateData = wu.get_worklist_updates(
            self.__worklist.worklist_id,
            body=self.__worklist.revision,
            filter=self.__worklist.wu_conf.worklist_filter,
        )

        if inc_updts is not None:
            updates: List[ClientWorklistItemUpdate] = []
            self.__iterate_updates(updates, inc_updts)
            self.__apply_worklist_updates(
                inc_updts.source_revision, inc_updts.target_revision, updates
            )
            self._notify_worklist_update_listeners(updates)
        return self.__items

    def __iterate_updates(
        self, updates: List[ClientWorklistItemUpdate], inc: IncWorklistUpdateData
    ):
        """Consumes the given worklist update iterator and appends all retrieved updates to the provided updates list."""
        if inc is None:
            return
        if inc.item_updates:
            updates += inc.item_updates
        else:
            return
        if inc.dropped:
            return

        # fetch next
        iwua: IncWorklistUpdateApi = self.__service_provider.get_service(IncWorklistUpdateApi)
        next_it: IncWorklistUpdateData = iwua.inc_wl_updt_get_next(inc.inc_upd_id)
        self.__iterate_updates(updates, next_it)

    def __apply_worklist_updates(
        self,
        source_revision: WorklistRevision,
        target_revision: int,
        updates: List[ClientWorklistItemUpdate],
    ):
        """Applies the provided worklist updates to self.__items. Checks the consistency to the source revision,
        and performs a full update if the state does not fit. Sets the new target revision in self.__worklist.
        """
        if (
            self.__worklist.revision.update_count != source_revision.update_count
            or self.__worklist.revision.initialisation_date != source_revision.initialisation_date
        ):
            # out of order update, clear and re-fetch everything
            self.__items.clear()
            self.__worklist = None
            self.get_worklist()
            return

        # print(f'Applying {len(updates)} updates')
        # print(updates)

        for update in updates:
            self.__apply_worklist_update(update)

        # remember the update count for the next update
        self.__worklist.revision.update_count = target_revision

    def __apply_worklist_update(self, update: ClientWorklistItemUpdate):
        """Applies the given update to __items"""
        update_type = update.update_type
        item = update.item
        if update_type == "CHANGED":
            self.__replace_or_add_item(item)
        elif update_type == "ADDED":
            self.__items += [item]
        elif update_type == "REMOVED":
            self.__remove_item(item)
        elif update_type == "ADDED_OR_CHANGED":
            self.__replace_or_add_item(item)
        elif update_type == "REMOVED_OR_NOTHING":
            self.__remove_item(item)
        else:
            raise RuntimeError("Unknown update type: " + update_type)

    def __replace_or_add_item(self, item: ClientWorklistItem):
        """Replaces or adds the given worklist item in self.__items"""
        # print('__replace_or_add_item: __items=', self.__items)
        for i in range(len(self.__items)):
            val = self.__items[i]
            if item.id == val.id:
                self.__items[i] = item
                return
        # not found above, append it
        self.__items.append(item)

    def __remove_item(self, item: ClientWorklistItem):
        """Removes the given worklist item from self.__items"""
        for val in self.__items:
            if item.id == val.id:
                self.__items.remove(val)
                return

    def find_item_by_id(self, item_id: str) -> ClientWorklistItem:
        """Finds a worklist item by its worklist item id. Returns none, if not in the worklist of the user."""
        # print(f'Finding item with id {item_id}')
        self.update_worklist()
        # print(self.__items)
        for item in self.__items:
            if item.id == item_id:
                # print('Found')
                return item
        return None



# Default Python Libraries
import base64
from typing import List, Type, TypeVar

# AristaFlow REST Libraries
import af_execution_manager
from af_execution_manager.api.activity_execution_control_api import ActivityExecutionControlApi
from af_org_model_manager.api.global_security_manager_api import GlobalSecurityManagerApi
from af_org_model_manager.models.auth_data_app_name import AuthDataAppName
from af_org_model_manager.models.auth_data_arbitrary import AuthDataArbitrary
from af_org_model_manager.models.auth_data_org_pos import AuthDataOrgPos
from af_org_model_manager.models.auth_data_password import AuthDataPassword
from af_org_model_manager.models.auth_data_user_name import AuthDataUserName
from af_org_model_manager.models.authentication_data import AuthenticationData
from af_org_model_manager.models.client_session_details import ClientSessionDetails
from af_org_model_manager.models.qualified_agent import QualifiedAgent
from af_remote_html_runtime_manager.api.runtime_manager_api import RuntimeManagerApi
from af_remote_html_runtime_manager.api.synchronous_activity_starting_api import (
    SynchronousActivityStartingApi,
)
from af_remote_html_runtime_manager.models.activity_rest_callback_data import (
    ActivityRestCallbackData,
)
from af_remote_html_runtime_manager.models.ebp_instance_reference import EbpInstanceReference
from af_remote_html_runtime_manager.models.gui_context import GuiContext
from af_worklist_manager.models.af_activity_reference import AfActivityReference
from af_worklist_manager.models.worklist_item import WorklistItem
from aristaflow.absence_service import AbsenceService
from aristaflow.delegation_service import DelegationService
from aristaflow.execution_history_service import ExecutionHistoryService
from aristaflow.html_gui_context import HtmlGuiContext
from aristaflow.org_model_service import OrgModelService
from aristaflow.process_service import ProcessService
from aristaflow.service_provider import ServiceProvider

from .configuration import Configuration
from .worklist_service import WorklistService


T = TypeVar("T")
D = TypeVar("D")


class AristaFlowClientService(object):
    """Client Services for accessing the AristaFlow BPM platform"""

    # the user session id this client service belongs to
    __user_session: str = None
    # authentication
    __csd: ClientSessionDetails = None
    # the AristaFlow configuration
    __af_conf: Configuration = None
    __service_provider: ServiceProvider = None
    __worklist_service: WorklistService = None
    __process_service: ProcessService = None
    __delegation_service: DelegationService = None
    __absence_service: AbsenceService = None
    __execution_history_service: ExecutionHistoryService = None
    __org_model_service: OrgModelService = None

    def __init__(
        self, configuration: Configuration, user_session: str, service_provider: ServiceProvider
    ):
        self.__af_conf = configuration
        self.__user_session = user_session
        self.__service_provider = service_provider

    def get_service(self, service_type: Type[T]) -> T:
        """
        Returns a service instance for the given service type, e.g.
        get_service(InstanceControlApi)
        @param service_type The class of the requested service.
        """
        return self.__service_provider.get_service(service_type)

    @property
    def client_session_details(self) -> ClientSessionDetails:
        """Returns the client session details of this service
        :return: The client session details of this service
        """
        return self.__csd

    @property
    def is_authenticated(self) -> bool:
        """Returns true, if this client service is already authenticated"""
        return self.__csd is not None

    def authenticate(self, username: str, password: str = None, org_pos_id: int = None):
        if self.__csd is not None:
            raise Exception("Already authenticated")

        auth_data: List[AuthenticationData] = [
            AuthDataUserName(username, sub_class="AuthDataUserName")
        ]
        if org_pos_id is not None:
            auth_data.append(AuthDataOrgPos(org_pos_id, sub_class="AuthDataOrgPos"))
        psk = self.__af_conf.pre_shared_key
        method: str
        # if a password was provided, use it
        if password:
            auth_data.append(AuthDataPassword(password=password, sub_class="AuthDataPassword"))
            method = "UTF-8_PASSWORD"
        # if PSK is configured, use that
        elif psk:
            # get the utf-8 bytes, encode them using base 64 and decode the resulting bytes using ASCII
            psk_encoded = base64.b64encode(bytes(psk, "UTF-8")).decode("ascii")
            auth_data.append(AuthDataArbitrary(data=psk_encoded, sub_class="AuthDataArbitrary"))
            method = "SHARED_UTF-8_KEY"
        else:
            raise Exception("No authentication method left")

        gsm: GlobalSecurityManagerApi = self.get_service(GlobalSecurityManagerApi)

        # use a provided application name
        if self.__af_conf.application_name:
            if org_pos_id is None:
                # if an application name is provided, an org position ID must be used as well
                # get the org positions
                agents: List[QualifiedAgent] = gsm.pre_authenticate_method(method, body=auth_data)
                agent: QualifiedAgent
                # pick the single org position
                if len(agents) == 1:
                    agent = agents[0]
                # none: can't login
                elif len(agents) == 0:
                    raise Exception(
                        f"User does not have an org position {username} (supplied org position id: {org_pos_id})"
                    )
                else:
                    # use the first org position, except there is a agent_name/username match
                    agent = agents[0]
                    for a in agents:
                        if a.agent_name == username:
                            agent = a
                            break
                # set the org position for the actual authentication
                auth_data.append(AuthDataOrgPos(agent.org_pos_id, sub_class="AuthDataOrgPos"))
            # use the application name
            auth_data.append(
                AuthDataAppName(
                    app_name=self.__af_conf.application_name, sub_class="AuthDataAppName"
                )
            )

        csds: List[ClientSessionDetails] = gsm.authenticate_all_method(
            method, self.__af_conf.caller_uri, body=auth_data
        )

        csd: ClientSessionDetails
        if len(csds) == 1:
            csd = csds[0]
        elif len(csds) == 0:
            raise Exception(
                f"User does not have an org position {username} (supplied org position id: {org_pos_id})"
            )
        else:
            # pick the first as default
            csd = csds[0]
            # pick the one where username and org position name are the same
            for entry in csds:
                if entry.agent.agent.org_pos_name == entry.agent.agent.agent_name:
                    csd = entry
                    break

        self.__service_provider.authenticated(csd)
        self.__csd = csd

    def is_html_activity(self, item: WorklistItem) -> bool:
        return (
            item.act_ref.gui_context_id == "HTMLContext"
            or item.act_ref.executable_component_name == "de.aristaflow.form.Form"
            or item.act_ref.executable_component_name == "de.aristaflow.form.GeneratedForm"
        )

    @property
    def worklist_service(self):
        if self.__worklist_service is None:
            self.__worklist_service = WorklistService(self.__service_provider, self.__af_conf)
        return self.__worklist_service

    @property
    def process_service(self):
        if self.__process_service is None:
            self.__process_service = ProcessService(self.__service_provider, self.__af_conf)
        return self.__process_service

    @property
    def delegation_service(self):
        if self.__delegation_service is None:
            self.__delegation_service = DelegationService(self.__service_provider)
        return self.__delegation_service

    @property
    def absence_service(self):
        if self.__absence_service is None:
            self.__absence_service = AbsenceService(self.__service_provider)
        return self.__absence_service

    @property
    def execution_history_service(self):
        if self.__execution_history_service is None:
            self.__execution_history_service = ExecutionHistoryService(self.__service_provider)
        return self.__execution_history_service

    @property
    def org_model_service(self):
        if self.__org_model_service is None:
            self.__org_model_service = OrgModelService(self.__service_provider)
        return self.__org_model_service

    def start_html_activity(self, item: WorklistItem, callback_url: str):
        """
        Starts the given HTML GUI worklist item using the Remote HTML Runtime Manager
        """
        if item is None:
            raise Exception("No worklist item provided")
        # accept user form and HTMLContext based activities
        if not (self.is_html_activity(item)):
            raise Exception(f"Not an HTML activity: {item.act_ref.gui_context_id}")
        if item.state == "STARTED":
            raise Exception("Item is already started")
        sas = self.get_html_activity_starting()
        gc: GuiContext
        ar: AfActivityReference = item.act_ref
        # print('Starting activity...')
        ebp_ir: EbpInstanceReference = EbpInstanceReference(
            ar.type,
            ar.instance_id,
            ar.instance_log_id,
            ar.base_template_id,
            ar.node_id,
            ar.node_iteration,
            ar.execution_manager_uris,
            ar.runtime_manager_uris,
        )
        # "AVAILABLE", "ASSIGNED", "STARTED", "SUSPENDED", "ENQUIRED"
        if item.state == "AVAILABLE" or item.state == "ASSIGNED":
            if callback_url is not None:
                cb_data = ActivityRestCallbackData(
                    sub_class="ActivityRestCallbackData",
                    notification_callback=callback_url,
                    activity=ebp_ir,
                )
                gc = sas.start_activity_callback(body=cb_data)
            else:
                gc = sas.start_activity(body=ebp_ir)
        else:
            if callback_url is not None:
                cb_data = ActivityRestCallbackData(
                    sub_class="ActivityRestCallbackData",
                    notification_callback=callback_url,
                    activity=ebp_ir,
                )
                gc = sas.resume_activity_callback(body=cb_data)
            else:
                gc = sas.resume_activity(body=ebp_ir)
        return HtmlGuiContext(gc)

    def get_html_activity_starting(self) -> SynchronousActivityStartingApi:
        """
        Returns the Remote HTML Runtime Manager Synchronous Activity Starting, ensuring logon to the Runtime Manager
        """
        sas: SynchronousActivityStartingApi = self.get_service(SynchronousActivityStartingApi)
        rm: RuntimeManagerApi = self.get_service(RuntimeManagerApi)
        # always logon again, since the server might have been restarted in the meantime
        rm.logon(body=self.__csd)
        return sas

    def reset_activity(self, item: WorklistItem):
        """Resets the given worklist item."""
        if item.state != "STARTED":
            # nothing to do
            return
        # TODO check for a "local" activity for a soft reset
        ar: AfActivityReference = item.act_ref
        ebp_ir = af_execution_manager.EbpInstanceReference(
            ar.type,
            ar.instance_id,
            ar.instance_log_id,
            ar.base_template_id,
            ar.node_id,
            ar.node_iteration,
            ar.execution_manager_uris,
            ar.runtime_manager_uris,
        )
        aec: ActivityExecutionControlApi = self.get_service(ActivityExecutionControlApi)
        aec.reset_to_prev_savepoint(body=ebp_ir, force=True)

    def deserialize(self, data, klass: Type[D]) -> D:
        """Deserialize data using the given class of the generated OpenAPI models."""
        return self.__service_provider.deserialize(data, klass)

    def serialize(self, obj) -> str:
        """Serialize REST model object"""
        return self.__service_provider.serialize(obj)

    @property
    def autostart_timeout_seconds(self) -> int:
        """Wait time in seconds for auto start signals"""
        return self.__af_conf.autostart_timeout_seconds



# AristaFlow REST Libraries
from af_org_model_manager.api.policy_resolution_api import PolicyResolutionApi
from af_org_model_manager.models.qa_remote_iterator_data import QaRemoteIteratorData
from aristaflow.abstract_service import AbstractService
from aristaflow.utils import OrgUtils


class OrgModelService(AbstractService):
    """
    Helper methods for accessing the organizational model
    """

    def get_qualified_agent(self, agent_id: int, org_pos_id: int):
        """
        Returns the qualified agent object for the given IDs
        """
        pr: PolicyResolutionApi = self._service_provider.get_service(PolicyResolutionApi)
        it: QaRemoteIteratorData = pr.resolve_policy(
            org_policy=OrgUtils.build_staff_assignment_rule_for_agent(agent_id, org_pos_id)
        )
        return it.agents[0]



# -*- coding: utf-8 -*-

# AristaFlow REST Libraries
from aristaflow.client_platform import AristaFlowClientPlatform
from aristaflow.client_service import AristaFlowClientService
from aristaflow.configuration import Configuration
from aristaflow.delegation_service import DelegationService
from aristaflow.process_service import ProcessService
from aristaflow.remote_iterator_handler import RemoteIteratorHandler
from aristaflow.service_provider import ServiceProvider
from aristaflow.worklist_service import WorklistService



# coding: utf-8

"""

"""

# Third Party Libraries
from setuptools import find_packages, setup  # noqa: H301


# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools

REQUIRES = ["urllib3 >= 1.15", "six >= 1.10", "certifi", "python-dateutil", "sseclient"]

setup(
    name="aristaflowpy",
    version="1.9.0",
    description="AristaFlow BPM",
    author_email="info@aristaflow.com",
    url="https://github.com/aristaflow/aristaflowpy",
    keywords=["AristaFlow", "BPM", "Workflow", "REST"],
    install_requires=REQUIRES,
    packages=find_packages(),
    include_package_data=True,
    long_description="""\
        AristaFlow BPM Python integration
        """,
)



class Configuration(object):
    # TODO remote (html) runtime manager
    # TODO PIR
    def __init__(
        self,
        base_url: str,
        caller_uri: str,
        verify_ssl=True,
        pre_shared_key: str = None,
        application_name: str = None,
        async_thread_pool_size: int = None,
        autostart_timeout_seconds=30,
    ):
        """
        :param baseUrl
        """
        if base_url.endswith("/"):
            base_url = base_url[0 : len(base_url) - 1]
        if not (base_url.endswith("AristaFlowREST")):
            base_url = base_url + "/AristaFlowREST"
        self.__baseUrl = base_url
        self.__caller_uri = caller_uri
        self.__verify_ssl = verify_ssl
        self.__pre_shared_key = pre_shared_key
        self.__application_name = application_name
        self.__async_thread_pool_size = async_thread_pool_size
        self.__autostart_timeout_seconds = autostart_timeout_seconds

    @property
    def baseUrl(self) -> str:
        return self.__baseUrl

    @property
    def caller_uri(self) -> str:
        return self.__caller_uri

    @property
    def verify_ssl(self) -> bool:
        return self.__verify_ssl

    @property
    def pre_shared_key(self) -> str:
        return self.__pre_shared_key

    @property
    def application_name(self) -> str:
        return self.__application_name

    @property
    def async_thread_pool_size(self) -> int:
        return self.__async_thread_pool_size

    @property
    def autostart_timeout_seconds(self) -> int:
        return self.__autostart_timeout_seconds

    @property
    def sse_connect_retry_wait(self) -> float:
        """
        Sleep time in seconds between connection retries on lost SSE connections.
        """
        return 5.0

    def get_host(self, service_type: str, service_instance: str = None) -> str:
        """
        Returns the host definition for the given service type / instance, based on the configuration.
        :param service_type: The BPM service type name, e.g. WorklistManager
        :param service_instance: Optionally the simple BPM service instance name, e.g. RemoteHTMLRuntimeManager
        :return: str The host value for the requested service
        """
        if service_instance is None:
            if service_type == "RuntimeManager":
                service_instance = "RemoteHTMLRuntimeManager"
            else:
                service_instance = service_type
        return self.baseUrl + "/" + service_type + "/" + service_instance

    def get_debug(self, service_type: str, service_instance: str = None) -> bool:
        return False
        # return service_type == "RuntimeManager"
